#!/usr/bin/env python3

"""
Tool for plotting a custom figure.
"""

import logging
from typing import Type, List, TypedDict
from pydantic import BaseModel, Field
import streamlit as st
from langchain_openai import ChatOpenAI
from langchain_core.tools import BaseTool
from langchain_core.prompts import ChatPromptTemplate

# Initialize logger
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

class CustomHeader(TypedDict):
    """
    A list of headers extracted from the user question.
    """
    y: List[str]

class CustomPlotterInput(BaseModel):
    """
    Input schema for the PlotImage tool.
    """
    question: str = Field(description="Description of the plot")

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class CustomPlotterTool(BaseTool):
    """
    Tool for plotting a custom plot.
    """
    name: str = "custom_plotter"
    description: str = "A tool to plot or visualize the simulation results."
    args_schema: Type[BaseModel] = CustomPlotterInput
    st_session_key: str = None

    def _run(self, question: str) -> str:
        """
        Run the tool.

        Args:
            question (str): The question to ask about the model description.

        Returns:
            str: The answer to the question
        """
        # Check if sys_bio_model is provided
        model_object = st.session_state[self.st_session_key]
        if model_object is None:
            return "Please run the simulation first before plotting the figure."
        if model_object.simulation_results is None:
            return "Please run the simulation first before plotting the figure."
        df = model_object.simulation_results
        species_names = "\n".join(df.columns.tolist())
        llm = ChatOpenAI(model="gpt-4o")
        system = f"""
        Given the user question: {question},
        and the species: {species_names},
        which species are relevant to the user's question?
        """
        llm_with_structured_output = llm.with_structured_output(CustomHeader)
        system_prompt_structured_output = ChatPromptTemplate.from_template(system)
        chain = system_prompt_structured_output | llm_with_structured_output
        results = chain.invoke({"input": question})
        logger.info("Suggestions: %s", results)
        extracted_species = []
        for species in results['y']:
            if species in df.columns.tolist():
                extracted_species.append(species)
        logger.info("Extracted species: %s", extracted_species)
        st.session_state.custom_simulation_results = extracted_species
        if len(extracted_species) == 0:
            return "No species found in the simulation results that matches the user prompt."
        return "Plotted the figure using the following species: " + str(extracted_species)



#!/usr/bin/env python3

"""
BasicoModel class for loading and simulating SBML models
using the basico package.
"""

from typing import Optional, Dict, Union
from time import sleep
from urllib.error import URLError
from pydantic import Field, model_validator
import pandas as pd
import basico
from .sys_bio_model import SysBioModel

class BasicoModel(SysBioModel):
    """
    Model that loads and simulates SBML models using the basico package.
    Can load models from an SBML file or download them using a BioModels model_id.
    """
    model_id: Optional[int] = Field(None, description="BioModels model ID to download and load")
    sbml_file_path: Optional[str] = Field(None, description="Path to an SBML file to load")
    simulation_results: Optional[str] = None
    name: Optional[str] = Field("", description="Name of the model")
    description: Optional[str] = Field("", description="Description of the model")

    # Additional attribute not included in the schema
    copasi_model: Optional[object] = None  # Holds the loaded Copasi model

    @model_validator(mode="after")
    def check_model_id_or_sbml_file_path(self):
        """
        Validate that either model_id or sbml_file_path is provided.
        """
        if not self.model_id and not self.sbml_file_path:
            raise ValueError("Either model_id or sbml_file_path must be provided.")
        if self.model_id:
            attempts = 0
            max_retries = 5
            while attempts < max_retries:
                try:
                    self.copasi_model = basico.load_biomodel(self.model_id)
                    break
                except URLError as e:
                    attempts += 1
                    sleep(10*attempts)
                    if attempts >= max_retries:
                        raise e
            self.description = basico.biomodels.get_model_info(self.model_id)["description"]
        elif self.sbml_file_path:
            self.copasi_model = basico.load_model(self.sbml_file_path)
        return self

    def simulate(self,
                 parameters: Optional[Dict[str, Union[float, int]]] = None,
                 duration: Union[int, float] = 10,
                 interval: int = 10
                 ) -> pd.DataFrame:
        """
        Simulate the COPASI model over a specified range of time points.
        
        Args:
            parameters: Dictionary of model parameters to update before simulation.
            duration: Duration of the simulation in time units.
            interval: Interval between time points in the simulation.
        
        Returns:
            Pandas DataFrame with time-course simulation results.
        """

        # Update parameters in the model
        if parameters:
            for param_name, param_value in parameters.items():
                # check if the param_name is not None
                if param_name is None:
                    continue
                # if param is a kinectic parameter
                df_all_params = basico.model_info.get_parameters(model=self.copasi_model)
                if param_name in df_all_params.index.tolist():
                    basico.model_info.set_parameters(name=param_name,
                                                exact=True,
                                                initial_value=param_value,
                                                model=self.copasi_model)
                # if param is a species
                else:
                    basico.model_info.set_species(name=param_name,
                                                exact=True,
                                                initial_concentration=param_value,
                                                model=self.copasi_model)

        # Run the simulation and return results
        df_result = basico.run_time_course(model=self.copasi_model,
                                        intervals=interval,
                                        duration=duration)
        df_result.columns = df_result.columns.str.replace('{', '[', regex=False).\
                    str.replace('}', ']', regex=False)
        df_result.reset_index(inplace=True)
        self.simulation_results = df_result
        return df_result.copy()

    def get_model_metadata(self) -> Dict[str, Union[str, int]]:
        """
        Retrieve metadata specific to the COPASI model.
        
        Returns:
            Dictionary of model metadata.
        """
        return {
            "Model Type": "SBML Model (COPASI)",
            "Parameter Count": len(basico.get_parameters())
        }



#!/usr/bin/env python3

"""
Tool for fetching species and parameters from the model.
"""

from urllib.error import URLError
from time import sleep
from typing import Type
import basico
from pydantic import BaseModel, Field
from langchain_core.tools import BaseTool
import streamlit as st

class FetchParametersInput(BaseModel):
    """
    Input schema for the ResetModel tool.
    """
    fetch_species: bool = Field(description="Fetch species from the model.")
    fetch_parameters: bool = Field(description="Fetch parameters from the model.")

# Note: It's important that every field has type hints. BaseTool is a
# Pydantic class and not having type hints can lead to unexpected behavior.
class FetchParametersTool(BaseTool):
    """
    This tool fetches species and parameters from the model 
    and returns them as a string in a dictionary.
    """
    name: str = "fetch_parameters"
    description: str = "A tool for fetching species and parameters from the model."
    args_schema: Type[BaseModel] = FetchParametersInput
    st_session_key: str = None

    def _run(self,
             fetch_species: bool,
             fetch_parameters: bool
             ) -> str:
        """
        Run the tool.

        Args:
            fetch_species (bool): Fetch species from the model.
            fetch_parameters (bool): Fetch parameters from the model.

        Returns:
            dict: The species and parameters from the model.
        """
        model_obj = st.session_state[self.st_session_key]
        # Extract species from the model
        species = []
        if fetch_species:
            df_species = self.get_species_and_parameters(model_obj, fetch_species=True)
            species = df_species.index.tolist()
            species = ','.join(species)

        # Extract parameters from the model
        parameters = []
        if fetch_parameters:
            df_parameters = self.get_species_and_parameters(model_obj, fetch_species=False)
            parameters = df_parameters.index.tolist()
            parameters = ','.join(parameters)
        return {'Species': species, 'Parameters': parameters}

    def get_species_and_parameters(self,
                                   model_obj,
                                   fetch_species: bool):
        """
        Get the species from the model.
        """
        attempts = 0
        max_retries = 3
        while attempts < max_retries:
            try:
                if fetch_species:
                    df = basico.model_info.get_species(model=model_obj.copasi_model)
                else:
                    df = basico.model_info.get_parameters(model=model_obj.copasi_model)
                break
            except URLError as e:
                attempts += 1
                sleep(10)
                if attempts >= max_retries:
                    raise e
        return df
